#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/pcpu.h>
#include <sys/proc.h>
#include <sys/kthread.h>

#include <vm/vm.h>
#include <vm/vm_extern.h>

#include <x86/segments.h>

#include <machine/pcb.h>

static char gsarea[2 * PAGE_SIZE] __aligned(PAGE_SIZE);
CTASSERT(sizeof(gsarea) >= 2 * PAGE_SIZE);

SYSCTL_NODE(_debug, OID_AUTO, inout, CTLFLAG_RW, 0, "");

static int port = 0xeee0;
SYSCTL_INT(_debug_inout, OID_AUTO, port, CTLFLAG_RW, &port, 0, NULL);

static int sysctl_count = sizeof(gsarea);
SYSCTL_INT(_debug_inout, OID_AUTO, count, CTLFLAG_RW, &sysctl_count, 0, NULL);

#define	COUNTUP		0
#define	COUNTDOWN	1
void my_outs(int port, int offset, int count, int direction, int size);
void my_insb(int port, int offset, int count);

static void
modify_gs(int gssel, struct segment_descriptor *gsdesc)
{

	critical_enter();
	curthread->td_pcb->pcb_gsd = *gsdesc;
	PCPU_GET(fsgs_gdt)[1] = *gsdesc;
	critical_exit();
	load_gs(gssel);
}

static void
create_desc(struct segment_descriptor *desc, const void *ptr, u_int size,
    u_int type, int def32, int granularity)
{
	u_int base, limit;

	base = (u_int)ptr;
	limit = size - 1;

	if (limit > 0xfffff)
		panic("create_desc: invalid limit %#x", limit);

	desc->sd_lobase = base & 0xffffff;
	desc->sd_hibase = (base >> 24) & 0xff;
	desc->sd_lolimit = limit & 0xffff;
	desc->sd_hilimit = (limit >> 16) & 0xf;
	desc->sd_type  = type;
	desc->sd_dpl   = SEL_KPL;
	desc->sd_p     = 1;
	desc->sd_xx    = 0;
	desc->sd_def32 = def32 ? 1 : 0;
	desc->sd_gran  = granularity;
}

static void
create_readonly_desc(struct segment_descriptor *d, const void *p, u_int s)
{
	create_desc(d, p, s, SDT_MEMROA, 1, 0);
}

static void
create_readwrite_desc(struct segment_descriptor *d, const void *p, u_int s)
{
	create_desc(d, p, s, SDT_MEMRWA, 1, 0);
}

static void
create_growdown_desc(struct segment_descriptor *d, const void *p, u_int s,
    int big, int granularity)
{
	create_desc(d, p, s, SDT_MEMRWDA, big, granularity);
}

static __inline void
load_es(u_short sel)
{       
	__asm __volatile("movw %0,%%es" : : "rm" (sel));
}

static int
sysctl_doit(SYSCTL_HANDLER_ARGS)
{
	int error, doit, origsel, newsel;
	int count, direction, ins, outs, size;
	u_int segoff, base, limit;
	struct segment_descriptor origdesc, newdesc;
	vm_offset_t kva;

	size = 1;
	count = sysctl_count;
	direction = COUNTUP;
	outs = ins = 0;
	segoff = 0;
	kva = 0;

	origsel = rgs();
	origdesc = curthread->td_pcb->pcb_gsd;

	doit = 0;
	error = sysctl_handle_int(oidp, &doit, 0, req);
	if (error || req->newptr == NULL || doit == 0)
		return (error);

	switch (doit) {
	case 1:
		/* Trigger a #GP due to segment being unusable */
		newsel = GSEL(GNULL_SEL, SEL_KPL);
		newdesc = origdesc;
		outs = 1;
		break;
	case 2:
		/*
		 * Create a segment that will trigger a #GP if 'sysctl_count'
		 * is beyond the segment limit.
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		create_readwrite_desc(&newdesc, gsarea, sizeof(gsarea));
		outs = 1;
		break;
	case 3:
		/* 
		 * Trigger a #PF by reading from an address that is not mapped.
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		kva = kva_alloc(PAGE_SIZE);
		create_readwrite_desc(&newdesc, (void *)kva, PAGE_SIZE);
		outs = 1;
		break;
	case 4:
		/*
		 * Trigger a #PF by trying to write to a readonly address.
		 *
		 * Since we don't want segmentation to get in the way we
		 * create a flat segment from 0 to 4GB.
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		create_desc(&newdesc, 0, 0xfffff, SDT_MEMRWA, 1, 1);
		ins = 1;
		segoff = (u_int)zero_region;
		break;
	case 5:
		/*
		 * Trigger a #GP by doing an insb() to a readonly segment.
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		create_readonly_desc(&newdesc, gsarea, sizeof(gsarea));
		ins = 1;
		break;
	case 6:
		/*
		 * Trigger a #GP by doing an outsb() beyond the limits of
		 * a expand-down data segment: B=1, G=1
		 *
		 * An expand-down data segment pretty much forces the
		 * use of G=1. With G=0 the segment limit is below 1MB
		 * and thus the segment size is humongous [limit + 1, 4GB).
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		base = (u_int)gsarea + sizeof(gsarea);
		limit = 0xffffffff - sizeof(gsarea);
		create_growdown_desc(&newdesc, (void *)base,
		    (limit + 1) >> 12, 1, 1);

		outs = 1;
		segoff = 0xffffffff;
		direction = COUNTDOWN;
		break;
	case 7:
		/*
		 * Trigger a #GP by doing an outsb() beyond the limits of
		 * an expand-down data segment: B=0, G=0
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		base = (u_int)gsarea + sizeof(gsarea) - 0x10000;
		limit = 0xffff - sizeof(gsarea);
		create_growdown_desc(&newdesc, (void *)base, limit + 1, 0, 0);

		outs = 1;
		segoff = 0xffff;
		direction = COUNTDOWN;
		break;
	case 8:
		/*
		 * Trigger a #GP by doing an outsb() beyond the limits of
		 * an expand-down data segment: B=0, G=1
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		base = (u_int)gsarea + sizeof(gsarea) - 0x10000;
		limit = 0xffff - sizeof(gsarea);
		create_growdown_desc(&newdesc, (void *)base,
		    (limit + 1) >> 12, 0, 1);

		outs = 1;
		segoff = 0xffff;
		direction = COUNTDOWN;
		break;
	case 9:
		/*
		 * Trigger a #GP by doing an outsb() beyond the limits of
		 * an expand-down data segment: B=1, G=0
		 *
		 * XXX this is really hard to implement properly because
		 * B=1,G=0 creates a huge segment and thus makes it impossible
		 * to test the edge conditions. The low limit is less than
		 * 0x100000 and the high limit is 0xffffffff.
		 */
		return (EINVAL);
	case 10:
		/*
		 * Trigger an outs that spans two pages.
		 */
		newsel = GSEL(GUGS_SEL, SEL_KPL);
		create_readwrite_desc(&newdesc, gsarea, sizeof(gsarea));

		/*
		 * Trigger 3 outsl instructions:
		 * - first entirely on the first page [4091, 4092, 4093, 4094]
		 * - second spanning the pages [4095, 0, 1, 2]
		 * - third entirely on the second page [3, 4, 5, 6]
		 */
		outs = 1;
		segoff = 4091;
		count = 3;
		size = 4;
		break;
	default:
		return (EINVAL);
	}

	if (outs) {
		modify_gs(newsel, &newdesc);
		my_outs(port, segoff, count, direction, size);
		modify_gs(origsel, &origdesc);
	}

	if (ins) {
		/*
		 * The insb instruction only works with %es segment. But
		 * this is not straightfoward because %es and %ds both point
		 * to the same segment descriptor in the GDT. Therefore we
		 * cannot modify the descriptor pointed to by %es because
		 * that would end up modifying the %ds segment as well.
		 *
		 * Instead, we modify %gs and then modify %es to point to it.
		 */
		modify_gs(newsel, &newdesc);		/* %gs -> newdesc */
		load_es(newsel);			/* %es -> newdesc */
		my_insb(port, segoff, count);
		load_es(GSEL(GDATA_SEL, SEL_KPL));	/* restore %es */
		modify_gs(origsel, &origdesc);		/* restore %gs */
	}

	if (kva)
		kva_free(kva, PAGE_SIZE);

	return (0);
}

SYSCTL_PROC(_debug_inout, OID_AUTO, doit, CTLTYPE_INT|CTLFLAG_RW, 0, 0, 
    sysctl_doit, "I", "");

static int
outsb_modevent(module_t mod, int type, void *unused)
{

        switch (type) {
        case MOD_LOAD:
		break;
	case MOD_UNLOAD:
		break;
	default:
		return (EINVAL);
	}

	return (0);
}

static moduledata_t outsb_mod_data= {
	"outsb",
	outsb_modevent,
};

DECLARE_MODULE(test, outsb_mod_data, SI_SUB_SMP + 1, SI_ORDER_ANY);
